/*
 * Linux Event Logging for the Enterprise
 * Copyright (c) International Business Machines Corp., 2001
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Please send e-mail to lkessler@users.sourceforge.net if you have
 *  questions or comments.
 *
 *  Project Website:  http://evlog.sourceforge.net/
 *
 */
 
#include <stdio.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include "callback.h"

#define MAX_BACKENDS	10
#define BACKENDS_DIR 	"/opt/evlog/plugins"
#define SMALL_BUF_SIZE	256

#ifdef DEBUG2
#define TRACE(fmt, args...)             fprintf(stderr, fmt, ##args)
#else
#define TRACE(fmt, args...)             /* fprintf(stderout, fmt, ##args) */
#endif

/* Type define the back end function pointer */
typedef int (* BE_FUNC_T) (const char * data1, const char * data2, evl_callback_t cb_func);
typedef char * (* BE_QUERYINF_T)();
typedef void (* BE_CALLBACKINF_T) (evl_callback_t cb_func);
typedef int (* BE_INITPLUGIN_T)();

static int be_add(void *plugin, BE_FUNC_T func);
static int be_release(BE_FUNC_T func);

typedef struct backend {
	void *plugin;
	BE_FUNC_T func;
	int state; 
} backend_t;

/* Backends table */
backend_t be_table[MAX_BACKENDS]; 
int be_cnt = 0;

/*
 * dlopens the shared lib whose name is name. Returns the handle returned by
 * dlopen().
 */
void * getPlugInHandle(char * name)
{
	void *hdle;
	char lib_path[256];
	char *ext;

	ext = (char *) strrchr(name, '.');

	/* Skip . and .. */
	if (!strcmp(name, ".") || !strcmp(name, ".."))
		return (void *)NULL;
	/* Skip if extension is not .so */
	if (ext == NULL || strcmp(ext, ".so"))
		return (void*)NULL;

	sprintf(lib_path, BACKENDS_DIR "/%s", name);	
	/* Load plugin */
	if (!(hdle = dlopen(lib_path, RTLD_LAZY))) {
		fprintf(stderr, "%s\n", dlerror());
		return (void *)NULL;
	}
	return hdle; 
}

BE_FUNC_T getPlugInInterface(void * handle)
{
	BE_FUNC_T func;
	BE_QUERYINF_T qi;
	BE_CALLBACKINF_T cbi;
	BE_INITPLUGIN_T initpi;
	char * func_name;
	int status;

	/* Get the back end interface function for this plugin */
	if (!(qi = dlsym(handle, "_evlQueryInterface"))) {
		fprintf(stderr, "%s\n", dlerror());
		return NULL;
	}
	func_name = (*qi)();
	/* 
	 * Initialize callback inf for backend so it can invoke some of
	 * evlogd functions.
	 */
	if (!(cbi = dlsym(handle, "_evlInitCallBack"))) {
		fprintf(stderr, "init callback : %s\n", dlerror());
		return NULL;
	}
	(*cbi)(evl_callback);
	
	if (!(initpi = dlsym(handle, "_evlInitPlugIn"))) {
		fprintf(stderr, "initialize plugIn : %s\n", dlerror());
		return NULL;
	}
	status = (*initpi)();
	if (!(func = dlsym(handle, func_name))) {
		free(func_name);
		fprintf(stderr, "%s\n", dlerror());
		return NULL;
	}

	if (status == -1) {
		dlclose(handle);
		return NULL;
	}

	free(func_name);
	return func;
}
/* Called during evlogd setup */
void 
be_init()
{
	struct dirent **plugins;
	int n;
	static void * handle;

	n = scandir(BACKENDS_DIR, &plugins, 0, 0);
	if (n < 0) {
		//perror("scandir");
	}
	else {
		while(n--) {
			BE_FUNC_T inf_func;
			
			handle = (void *)getPlugInHandle((char *)(plugins[n]->d_name));
			free(plugins[n]);
			if (!handle)
				continue;
			inf_func = getPlugInInterface(handle);
			if (!inf_func)
				continue;
			
			be_add(handle, inf_func);
		}
		free(plugins);
	}
}

void
be_cleanup()
{
	int i;
	TRACE("be_cleanup invoked...\n");
	for (i = 0; i < MAX_BACKENDS; i++) {
		if (be_table[i].plugin != NULL) {
			dlclose(be_table[i].plugin);
			TRACE("be_cleanup plugin index #%d is set to NULL...\n", i);
			be_table[i].plugin=NULL;
			be_table[i].func=NULL;
		}
	}
}

int
be_add(void *plugin, BE_FUNC_T func)
{
	int i;
	if (be_cnt >= MAX_BACKENDS) {
		fprintf(stderr, "No more room for your plugin.\n");
		return -1;
	}

	for (i = 0; i < MAX_BACKENDS; i++) {
		if (be_table[i].func == NULL) {
			be_table[i].func = func;	
			be_table[i].plugin = plugin;
			be_cnt++;
			break;
		}
	}
	return 0;
}

int
be_release(BE_FUNC_T func)
{
	int i;
	for (i = 0; i < MAX_BACKENDS; i++) {
		if (be_table[i].func == func) {
			be_table[i].func = NULL;
			dlclose(be_table[i].plugin);
			be_table[i].plugin = NULL;
			be_cnt--;
			break;
		}
	}
	return 0;
}

/*
 * Execute all registered functions in the function table
 * (also pass its arguments to those funtions).
 */ 
void 
be_run(const char *data1, const char *data2, evl_callback_t callback)
{
	int i;
	int func_ret;

	for(i = 0; i < MAX_BACKENDS; i++) {
		if (be_table[i].func != NULL) {
			func_ret = (*(be_table[i].func))(data1, data2, evl_callback);
			if (func_ret != 0) {
				/* backend function fail */
				be_release(be_table[i].func);
			}
		}
	}
}

#if 0
int 
main(char *argv[], int argc)
{
	/* Initialize backends */
	be_init();
	/* Pass data to back end functions, and run them */
	be_run("data1 is here", "data2 is also here", evl_callback);
	printf("\nclean up...");
	be_cleanup();
	printf("done.\n");
	return 0;
}

#endif

